/* Copyright (C) 2005  David Decotigny

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA. 
*/

#include <hwcore/paging.h>
#include <sos/thread.h>
#include <sos/assert.h>

#include "uaccess.h"


/**
 * Helper function to force the MMU to switch to the address space of
 * the process
 */
static sos_ret_t bindto_user_mm_context()
{
  struct sos_thread * current_thread;
  struct sos_mm_context * process_mm_ctxt;

  current_thread = sos_thread_get_current();
  SOS_ASSERT_FATAL(NULL != current_thread->process);

  /* Switch to the user's address space */
  process_mm_ctxt
    = sos_process_get_mm_context(current_thread->process);
  return sos_thread_change_current_mm_context(process_mm_ctxt);
}


/**
 * Helper macro to restore the MMU configuration prior to
 * bindto_user_mm_context()
 */
#define unbindto_user_mm_context() \
  sos_thread_change_current_mm_context(NULL)


sos_ret_t sos_copy_from_user(sos_vaddr_t kernel_to,
			     sos_uaddr_t user_from,
			     sos_size_t size)
{
  sos_ret_t retval;
  retval = bindto_user_mm_context();
  if (SOS_OK != retval)
    return retval;

  /* Make sure user is trying to access user space */
  if (user_from < SOS_PAGING_BASE_USER_ADDRESS)
    {
      unbindto_user_mm_context();
      return -SOS_EPERM;
    }
  if (user_from > SOS_PAGING_TOP_USER_ADDRESS - size)
    {
      unbindto_user_mm_context();
      return -SOS_EPERM;
    }

  memcpy((void*) kernel_to, (void*) user_from, size);

  retval = unbindto_user_mm_context();
  if (SOS_OK != retval)
    return retval;

  return size;
}


sos_ret_t sos_copy_to_user(sos_uaddr_t user_to,
			   sos_vaddr_t kernel_from,
			   sos_size_t size)
{
  sos_ret_t retval;
  retval = bindto_user_mm_context();
  if (SOS_OK != retval)
    return retval;

  /* Make sure user is trying to access user space */
  if (user_to < SOS_PAGING_BASE_USER_ADDRESS)
    {
      unbindto_user_mm_context();
      return -SOS_EPERM;
    }
  if (user_to > SOS_PAGING_TOP_USER_ADDRESS - size)
    {
      unbindto_user_mm_context();
      return -SOS_EPERM;
    }

  memcpy((void*) user_to, (void*) kernel_from, size);

  retval = unbindto_user_mm_context();
  if (SOS_OK != retval)
    return retval;

  return size;
}
